#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef _MPI
	#include <mpi.h>
#endif
#include "lis.h"
#include "lis_matvec.h"

LIS_MATVEC_CRS lis_matvec_crs_xxx[5][8] = { 
	{
		lis_matvec_crs   , lis_matvec_crs_u2,lis_matvec_crs_u3 ,lis_matvec_crs_u4,
		lis_matvec_crs_u5, lis_matvec_crs_u6,lis_matvec_crs_u7 ,lis_matvec_crs_u8,
	},
	{
		lis_matvec_crs_p1,lis_matvec_crs_p2 ,lis_matvec_crs_p3 ,lis_matvec_crs_p4,
		lis_matvec_crs_p5,lis_matvec_crs_p6 ,lis_matvec_crs_p7 ,lis_matvec_crs_p8,
	},
	{
		lis_matvec_dia_u1s, lis_matvec_dia_u2s,lis_matvec_dia_u3s ,lis_matvec_dia_u4s,
		lis_matvec_dia_u5s, lis_matvec_dia_u6s,lis_matvec_dia_u7s ,lis_matvec_dia_u8s,
	},
	{
		lis_matvec_dia_p1s,lis_matvec_dia_p2s ,lis_matvec_dia_p3s ,lis_matvec_dia_p4s,
		lis_matvec_dia_p5s,lis_matvec_dia_p6s ,lis_matvec_dia_p7s ,lis_matvec_dia_p8s,
	},
	{
		/*
		lis_matvec_dia_u4s_sse2,lis_matvec_dia_u4s_sse2 ,lis_matvec_dia_u4s_sse2 ,lis_matvec_dia_u4s_sse2,
		lis_matvec_dia_u6s_sse2,lis_matvec_dia_u6s_sse2 ,lis_matvec_dia_u8s_sse2 ,lis_matvec_dia_u8s_sse2,
		*/
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	}
};

/************************/
/* Unrolling            */
/************************/
void lis_matvec_crs_u2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2;
	LIS_SCALAR	t1,t2;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = 0.0;
		for(j=js+1;j<je;j+=2)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];

			j2  = *(jj++);
			t2 += *(vv++) * x[j2];
		}
		for(j-=1;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2;
	}

}

void lis_matvec_crs_u3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3;
	LIS_SCALAR	t1,t2,t3;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = 0.0;
		for(j=js+2;j<je;j+=3)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];

			j2  = *(jj++);
			t2 += *(vv++) * x[j2];

			j3  = *(jj++);
			t3 += *(vv++) * x[j3];
		}
		for(j-=2;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3;
	}

}

void lis_matvec_crs_u4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4;
	LIS_SCALAR	t1,t2,t3,t4;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = 0.0;
		for(j=js+3;j<je;j+=4)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];

			j2  = *(jj++);
			t2 += *(vv++) * x[j2];

			j3  = *(jj++);
			t3 += *(vv++) * x[j3];

			j4  = *(jj++);
			t4 += *(vv++) * x[j4];
		}
		for(j-=3;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4;
	}

}

void lis_matvec_crs_u5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5;
	LIS_SCALAR	t1,t2,t3,t4,t5;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = 0.0;
		for(j=js+4;j<je;j+=5)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];

			j2  = *(jj++);
			t2 += *(vv++) * x[j2];

			j3  = *(jj++);
			t3 += *(vv++) * x[j3];

			j4  = *(jj++);
			t4 += *(vv++) * x[j4];

			j5  = *(jj++);
			t5 += *(vv++) * x[j5];
		}
		for(j-=4;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5;
	}

}

void lis_matvec_crs_u6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	t1,t2,t3,t4,t5,t6;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = t6 = 0.0;
		for(j=js+5;j<je;j+=6)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];

			j2  = *(jj++);
			t2 += *(vv++) * x[j2];

			j3  = *(jj++);
			t3 += *(vv++) * x[j3];

			j4  = *(jj++);
			t4 += *(vv++) * x[j4];

			j5  = *(jj++);
			t5 += *(vv++) * x[j5];

			j6  = *(jj++);
			t6 += *(vv++) * x[j6];
		}
		for(j-=5;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5+t6;
	}

}

void lis_matvec_crs_u7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	t1,t2,t3,t4,t5,t6,t7;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = t6 = t7 = 0.0;
		for(j=js+6;j<je;j+=7)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];

			j2  = *(jj++);
			t2 += *(vv++) * x[j2];

			j3  = *(jj++);
			t3 += *(vv++) * x[j3];

			j4  = *(jj++);
			t4 += *(vv++) * x[j4];

			j5  = *(jj++);
			t5 += *(vv++) * x[j5];

			j6  = *(jj++);
			t6 += *(vv++) * x[j6];

			j7  = *(jj++);
			t7 += *(vv++) * x[j7];
		}
		for(j-=6;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5+t6+t7;
	}

}

void lis_matvec_crs_u8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5,j6,j7,j8;
	LIS_SCALAR	t1,t2,t3,t4,t5,t6,t7,t8;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = t6 = t7 = t8 = 0.0;
		for(j=js+7;j<je;j+=8)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];

			j2  = *(jj++);
			t2 += *(vv++) * x[j2];

			j3  = *(jj++);
			t3 += *(vv++) * x[j3];

			j4  = *(jj++);
			t4 += *(vv++) * x[j4];

			j5  = *(jj++);
			t5 += *(vv++) * x[j5];

			j6  = *(jj++);
			t6 += *(vv++) * x[j6];

			j7  = *(jj++);
			t7 += *(vv++) * x[j7];

			j8  = *(jj++);
			t8 += *(vv++) * x[j8];
		}
		for(j-=7;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5+t6+t7+t8;
	}

}

/************************/
/* Pipeline             */
/************************/
void lis_matvec_crs_p1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2;
	LIS_SCALAR	t1,t2;
	LIS_SCALAR	v1,v2;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = 0;
		if( js<je )
		{
			j1 = *(jj++); v1 = *(vv++);
			for(j=js+1;j<je;j++)
			{
				j2 = *(jj++); v2 = *(vv++);
				t1 += v1 * x[j1];

				j1 = j2; v1 = v2;
			}
			t1 += v1 * x[j1];
		}
		y[i] = t1;
	}
}

void lis_matvec_crs_p2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2;
	LIS_SCALAR	t1,t2;
	LIS_SCALAR	x1,x2;
	LIS_SCALAR	v1,v2;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = 0;
		if( je-js<3 )
		{
			j = js;
		}
		else
		{
			j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
			for(j=js+2;j<je;j+=2)
			{
				j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
				t1 += v1 * x1;

				j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
				t2 += v2 * x2;
			}
			t1 += v1 * x1;
			j -= 1;
		}
		for(;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2;
	}

}

void lis_matvec_crs_p3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3;
	LIS_SCALAR	t1,t2,t3;
	LIS_SCALAR	x1,x2,x3;
	LIS_SCALAR	v1,v2,v3;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = 0;
		if( je-js<4 )
		{
			j = js;
		}
		else
		{
			j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
			j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
			for(j=js+4;j<je;j+=3)
			{
				j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
				t1 += v1 * x1;

				j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
				t2 += v2 * x2;

				j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
				t3 += v3 * x3;
			}
			t1 += v1 * x1;
			t2 += v2 * x2;
			j -= 2;
		}
		for(;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3;
	}

}

void lis_matvec_crs_p4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4;
	LIS_SCALAR	t1,t2,t3,t4;
	LIS_SCALAR	x1,x2,x3,x4;
	LIS_SCALAR	v1,v2,v3,v4;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = 0;
		if( je-js<5 )
		{
			j = js;
		}
		else
		{
			j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
			j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
			j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
			for(j=js+6;j<je;j+=4)
			{
				j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
				t1 += v1 * x1;

				j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
				t2 += v2 * x2;

				j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
				t3 += v3 * x3;

				j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
				t4 += v4 * x4;
			}
			t1 += v1 * x1;
			t2 += v2 * x2;
			t3 += v3 * x3;
			j -= 3;
		}
		for(;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4;
	}

}

void lis_matvec_crs_p5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5;
	LIS_SCALAR	t1,t2,t3,t4,t5;
	LIS_SCALAR	x1,x2,x3,x4,x5;
	LIS_SCALAR	v1,v2,v3,v4,v5;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = 0;
		if( je-js<6 )
		{
			j = js;
		}
		else
		{
			j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
			j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
			j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
			j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
			for(j=js+8;j<je;j+=5)
			{
				j5 = *(jj++); v5 = *(vv++); x5 = x[j5];
				t1 += v1 * x1;

				j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
				t2 += v2 * x2;

				j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
				t3 += v3 * x3;

				j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
				t4 += v4 * x4;

				j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
				t5 += v5 * x5;
			}
			t1 += v1 * x1;
			t2 += v2 * x2;
			t3 += v3 * x3;
			t4 += v4 * x4;
			j -= 4;
		}
		for(;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5;
	}

}

void lis_matvec_crs_p6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	t1,t2,t3,t4,t5,t6;
	LIS_SCALAR	x1,x2,x3,x4,x5,x6;
	LIS_SCALAR	v1,v2,v3,v4,v5,v6;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = t6 = 0;
		if( je-js<7 )
		{
			j = js;
		}
		else
		{
			j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
			j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
			j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
			j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
			j5 = *(jj++); v5 = *(vv++); x5 = x[j5];
			for(j=js+10;j<je;j+=6)
			{
				j6 = *(jj++); v6 = *(vv++); x6 = x[j6];
				t1 += v1 * x1;

				j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
				t2 += v2 * x2;

				j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
				t3 += v3 * x3;

				j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
				t4 += v4 * x4;

				j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
				t5 += v5 * x5;

				j5 = *(jj++); v5 = *(vv++); x5 = x[j5];
				t6 += v6 * x6;
			}
			t1 += v1 * x1;
			t2 += v2 * x2;
			t3 += v3 * x3;
			t4 += v4 * x4;
			t5 += v5 * x5;
			j -= 5;
		}
		for(;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5+t6;
	}

}

void lis_matvec_crs_p7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	t1,t2,t3,t4,t5,t6,t7;
	LIS_SCALAR	x1,x2,x3,x4,x5,x6,x7;
	LIS_SCALAR	v1,v2,v3,v4,v5,v6,v7;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = t6 = t7 = 0;
		if( je-js<8 )
		{
			j = js;
		}
		else
		{
			j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
			j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
			j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
			j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
			j5 = *(jj++); v5 = *(vv++); x5 = x[j5];
			j6 = *(jj++); v6 = *(vv++); x6 = x[j6];
			for(j=js+12;j<je;j+=7)
			{
				j7 = *(jj++); v7 = *(vv++); x7 = x[j7];
				t1 += v1 * x1;

				j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
				t2 += v2 * x2;

				j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
				t3 += v3 * x3;

				j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
				t4 += v4 * x4;

				j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
				t5 += v5 * x5;

				j5 = *(jj++); v5 = *(vv++); x5 = x[j5];
				t6 += v6 * x6;

				j6 = *(jj++); v6 = *(vv++); x6 = x[j6];
				t7 += v7 * x7;
			}
			t1 += v1 * x1;
			t2 += v2 * x2;
			t3 += v3 * x3;
			t4 += v4 * x4;
			t5 += v5 * x5;
			t6 += v6 * x6;
			j -= 6;
		}
		for(;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5+t6+t7;
	}

}

void lis_matvec_crs_p8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j1,j2,j3,j4,j5,j6,j7,j8;
	LIS_SCALAR	t1,t2,t3,t4,t5,t6,t7,t8;
	LIS_SCALAR	x1,x2,x3,x4,x5,x6,x7,x8;
	LIS_SCALAR	v1,v2,v3,v4,v5,v6,v7,v8;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t1 = t2 = t3 = t4 = t5 = t6 = t7 = t8 = 0;
		if( je-js<9 )
		{
			j = js;
		}
		else
		{
			j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
			j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
			j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
			j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
			j5 = *(jj++); v5 = *(vv++); x5 = x[j5];
			j6 = *(jj++); v6 = *(vv++); x6 = x[j6];
			j7 = *(jj++); v7 = *(vv++); x7 = x[j7];
			for(j=js+14;j<je;j+=8)
			{
				j8 = *(jj++); v8 = *(vv++); x8 = x[j8];
				t1 += v1 * x1;

				j1 = *(jj++); v1 = *(vv++); x1 = x[j1];
				t2 += v2 * x2;

				j2 = *(jj++); v2 = *(vv++); x2 = x[j2];
				t3 += v3 * x3;

				j3 = *(jj++); v3 = *(vv++); x3 = x[j3];
				t4 += v4 * x4;

				j4 = *(jj++); v4 = *(vv++); x4 = x[j4];
				t5 += v5 * x5;

				j5 = *(jj++); v5 = *(vv++); x5 = x[j5];
				t6 += v6 * x6;

				j6 = *(jj++); v6 = *(vv++); x6 = x[j6];
				t7 += v7 * x7;

				j7 = *(jj++); v7 = *(vv++); x7 = x[j7];
				t8 += v8 * x8;
			}
			t1 += v1 * x1;
			t2 += v2 * x2;
			t3 += v3 * x3;
			t4 += v4 * x4;
			t5 += v5 * x5;
			t6 += v6 * x6;
			t7 += v7 * x7;
			j -= 7;
		}
		for(;j<je;j++)
		{
			j1  = *(jj++);
			t1 += *(vv++) * x[j1];
		}
		y[i] = t1+t2+t3+t4+t5+t6+t7+t8;
	}

}
