#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lis.h"
#include "lis_precon.h"

int lis_precon_create_ilu0(LIS_PRECON precon)
{
#ifdef _OPENMP
	int	i,j,jj,k,kk,n,gn,nnz,ndz;
	int is,ie,my_rank,nprocs;
	int count,jw,err;
	int *index, *iw, *d, *ju, *iw2;
	LIS_SCALAR t,w;
	LIS_SCALAR *value;
	LIS_MATRIX P, A;

	A      = precon->Amat;
	n      = A->n;
	gn     = A->gn;
	nnz    = A->nnz;
	w      = precon->params[LIS_PARAMS_RELAX];
	nprocs = omp_get_max_threads();

	index = NULL;
	value = NULL;
	iw    = NULL;
	d     = NULL;
	ju    = NULL;

	err = lis_matrix_duplicate(A,&P);
	if( err )
	{
		return err;
	}
	iw  = (int *)lis_malloc( (gn+1)*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}
	d   = (int *)lis_malloc( n*sizeof(int) );
	if( d==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}
	ju  = (int *)lis_malloc( n*sizeof(int) );
	if( ju==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}

	count = 0;
#pragma omp parallel private(i,j)
{
#pragma omp for
	for(i=0;i<n;i++)
	{
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
		iw[i+1] = 0;
		d[i]    = A->ptr[i+1];
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				iw[i+1] = 1;
				d[i]    = j;
			}
		}
	}
#pragma omp for reduction(+:count)
	for(i=0;i<n;i++)
	{
		count += iw[i+1];
	}
#pragma omp for
	for(i=0;i<n;i++)
	{
		iw[i+1] = A->ptr[i+1]-A->ptr[i]-iw[i+1];
	}
}
	ndz = n - count;

	index = (int *)lis_malloc((nnz+ndz+1)*sizeof(int));
	if( index==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}
	value = (LIS_SCALAR *)lis_malloc((nnz+ndz+1)*sizeof(LIS_SCALAR));
	if( value==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}

	iw[0] = n+1;
	for(i=0;i<n;i++)
	{
		iw[i+1] = iw[i+1] + iw[i];
	}
#pragma omp parallel private(i,j,jj,is,ie,k,kk,my_rank,t,jw,iw2,w)
{
	iw2     = (int *)lis_malloc( n*sizeof(int) );
	w       = precon->params[LIS_PARAMS_RELAX];
	my_rank = omp_get_thread_num();
	LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

	for(i=is;i<ie;i++)
	{
		index[i] = iw[i];
	}
	#pragma omp single
	index[n] = iw[n];
	for(i=0;i<n;i++)
	{
		iw2[i] = -1;
	}

	for(i=is;i<ie;i++)
	{
		value[i] = A->value[d[i]];
	}
	
	for(i=is;i<ie;i++)
	{
		k = index[i];
		iw2[i] = i;
		for(j=A->ptr[i];j<d[i];j++)
		{
			value[k] = A->value[j];
			index[k] = A->index[j];
			iw2[A->index[j]] = k;
			k++;
		}
		ju[i] = k;
		for(j=d[i]+1;j<A->ptr[i+1];j++)
		{
			value[k] = A->value[j];
			index[k] = A->index[j];
			iw2[A->index[j]] = k;
			k++;
		}
		for(j=index[i];j<ju[i];j++)
		{
			jj          = index[j];
			if( jj<is ) continue;
			t           = value[j] * value[jj];
			value[j] = t;

			for(k=ju[jj];k<index[jj+1];k++)
			{
				kk = index[k];
				if( kk<is || kk>=ie ) continue;
				jw = iw2[kk];
				if(jw!=-1)
				{
					value[jw] -= w * t * value[k];
				}
			}
		}
		value[i] = 1.0 / value[i];

		for(j=A->ptr[i];j<A->ptr[i+1];j++) iw2[A->index[j]] = -1;
	}
	lis_free(iw2);
}
	lis_matrix_set_msr(nnz,ndz,index,value,P);
	lis_matrix_assemble(&P,LIS_MATRIX_MSR);
	lis_free(iw);
	lis_free(d);

	P->ptr       = ju;
	precon->Pmat = P;
	return LIS_SUCCESS;
#else
	int	i,j,jj,k,kk,n,gn,nnz,ndz;
	int count,jw,err;
	int *iw, *d, *ju;
	int *index;
	LIS_SCALAR t,w;
	LIS_SCALAR *value;
	LIS_MATRIX P, A;
	LIS_MATRIX_PARAMS params;

	A   = precon->Amat;
	n   = A->n;
	gn  = A->gn;
	nnz = A->nnz;
	w   = precon->params[LIS_PARAMS_RELAX];

	index = NULL;
	value = NULL;
	iw    = NULL;
	d     = NULL;
	ju    = NULL;

	err = lis_matrix_duplicate(A,&P);
	if( err )
	{
		return err;
	}
	iw  = (int *)lis_malloc( (gn+1)*sizeof(int) );
	if( iw==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}
	d   = (int *)lis_malloc( n*sizeof(int) );
	if( d==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}
	ju  = (int *)lis_malloc( n*sizeof(int) );
	if( ju==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}

	count = 0;
	for(i=0;i<n;i++)
	{
#ifdef _MPI
		lis_sort_iid(A->ptr[i],A->ptr[i+1]-1,A->index,A->g_index,A->value);
#else
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
#endif
		iw[i+1] = 0;
		d[i]    = A->ptr[i+1];
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				count++;
				iw[i+1] = 1;
				d[i]    = j;
			}
		}
	}
	for(i=0;i<n;i++)
	{
		iw[i+1] = A->ptr[i+1]-A->ptr[i]-iw[i+1];
	}
	ndz = n - count;

	index = (int *)lis_malloc((nnz+ndz+1)*sizeof(int));
	if( index==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}
	value = (LIS_SCALAR *)lis_malloc((nnz+ndz+1)*sizeof(LIS_SCALAR));
	if( value==NULL )
	{
		lis_free2(5,index,value,iw,d,ju);
		return LIS_OUT_OF_MEMORY;
	}

	iw[0] = n+1;
	for(i=0;i<n;i++)
	{
		iw[i+1] = iw[i+1] + iw[i];
	}

	for(i=0;i<=n;i++)
	{
		index[i] = iw[i];
		iw[i] = -1;
	}

	for(i=0;i<n;i++)
	{
		value[i] = A->value[d[i]];
	}
	for(i=0;i<n;i++)
	{
		k = index[i];
		iw[i] = i;
		for(j=A->ptr[i];j<d[i];j++)
		{
			jj = A->index[j];
			value[k] = A->value[j];
			index[k] = A->index[j];
			iw[jj] = k;
			k++;
		}
		ju[i] = k;
		for(j=d[i]+1;j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
//			if( jj>=n ) continue;
			value[k] = A->value[j];
			index[k] = A->index[j];
			iw[jj] = k;
			k++;
		}

		for(j=index[i];j<ju[i];j++)
		{
			jj          = index[j];
			if( jj>=n ) continue;
			t           = value[j] * value[jj];
			value[j] = t;

			for(k=ju[jj];k<index[jj+1];k++)
			{
				kk = index[k];
				if( kk>=n ) continue;
				jw = iw[kk];
				if(jw!=-1) value[jw] -= w * t * value[k];
			}
		}
		value[i] = 1.0 / value[i];

		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			jj = A->index[j];
			iw[jj] = -1;
		}
	}
	lis_matrix_set_msr(nnz,ndz,index,value,P);
	lis_matrix_assemble(&P,LIS_MATRIX_MSR);
	lis_free(iw);
	lis_free(d);

	P->ptr       = ju;
	precon->Pmat = P;
	return LIS_SUCCESS;
#endif
}


void lis_psolve_ilu0(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
#ifdef _OPENMP
	int i,j,n;
	int k,jj,is,ie,threads;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;

	/*
	 *  LUx = b
	 *  LU  = (D + L*A) * (I + D^-1 * U*A)
	 */

	n       = precon->Pmat->n;
	b       = B->value;
	x       = X->value;
	threads = omp_get_max_threads();
#pragma omp parallel private(i,j,jj,is,ie,k,t)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);

	for(i=is; i<ie; i++)
	{
		t = b[i];
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t;
	}
	for(i=ie-1; i>=is; i--)
	{
		t = x[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t * precon->Pmat->value[i];
	}
}
#else
	int i,j,jj,n;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;

	/*
	 *  LUx = b
	 */

	n = precon->Pmat->n;
	b = B->value;
	x = X->value;
	for(i=0; i<n; i++)
	{
		t = b[i];
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj<n ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t;
	}
	for(i=n-1; i>=0; i--)
	{
		t = x[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj<n ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t * precon->Pmat->value[i];
	}
#endif
}

void lis_psolvet_ilu0(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
#ifdef _OPENMP
	int i,j,n;
	int k,jj,is,ie,threads;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;

	/*
	 *  (LU)'x = b
	 *  LU     = (D + L*A) * (I + D^-1 * U*A)
	 */

	n       = precon->Pmat->n;
	b       = B->value;
	x       = X->value;
	threads = omp_get_max_threads();
#pragma omp parallel private(i,j,jj,is,ie,k,t)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);

	for(i=is; i<ie; i++)
	{
		x[i] = b[i];
	}
	for(i=is; i<ie; i++)
	{
		x[i] = x[i] * precon->Pmat->value[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) x[jj] -= precon->Pmat->value[j] * x[i];
		}
	}
	for(i=ie-1; i>=is; i--)
	{
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) x[jj] -= precon->Pmat->value[j] * x[i];
		}
	}
}
#else
	int i,j,n;
	LIS_SCALAR *b,*x;

	/*
	 *  (LU)'x = b
	 */

	n = precon->Pmat->n;
	b = B->value;
	x = B->value;
	for(i=0; i<n; i++)
	{
		x[i] = b[i];
	}
	for(i=0; i<n; i++)
	{
		x[i] = x[i] * precon->Pmat->value[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			x[precon->Pmat->index[j]] -= precon->Pmat->value[j] * x[i];
		}
	}
	for(i=n-1; i>=0; i--)
	{
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			x[precon->Pmat->index[j]] -= precon->Pmat->value[j] * x[i];
		}
	}
#endif
}

